import pprint


class DebugTask:
    def __init__(self, args, project):
        self.args = args
        self.project = project

    def run(self):
        print("args: {}".format(self.args))
        print("project: ")
        pprint.pprint(self.project)



import pprint
import os
import fnmatch
import jinja2
import yaml
import dbt.project
from collections import defaultdict

from dbt.compilation import Linker

CREATE_STATEMENT_TEMPLATE = """
create {table_or_view} {schema}.{identifier} {dist_qualifier} {sort_qualifier} as (
    {query}
);"""

class CompileTask:
    def __init__(self, args, project):
        self.args = args
        self.project = project

    def __project_sources(self, project):
        """returns: {'model': ['pardot/model.sql', 'segment/model.sql']}
        """
        indexed_files = defaultdict(list)

        for source_path in project['source-paths']:
            full_source_path = os.path.join(project['project-root'], source_path)
            for root, dirs, files in os.walk(full_source_path):
                for filename in files:
                    abs_path = os.path.join(root, filename)
                    rel_path = os.path.relpath(abs_path, full_source_path)

                    if fnmatch.fnmatch(filename, "*.sql"):
                        indexed_files[full_source_path].append((project, rel_path))

        return indexed_files

    def __project_models(self, project_sources):
        project_models = []
        for (source_path, model_definitions) in project_sources.items():

            for (project, model_file) in model_definitions:

                package_name = project.cfg['name']

                filename = os.path.basename(model_file)
                model_group = os.path.dirname(model_file)
                model_name  = os.path.splitext(filename)[0]
                model = (package_name, model_group, model_name)

                if model in project_models:
                    print("WARNING: Conflicting model found package={}, model={}".format(package_name, model_name))

                # add the conflict and catch it if it's used in compilation
                project_models.append(model)

        return project_models


    def __write(self, path, payload):
        target_path = os.path.join(self.project['target-path'], path)

        if not os.path.exists(os.path.dirname(target_path)):
            os.makedirs(os.path.dirname(target_path))
        elif os.path.exists(target_path):
            print("Compiler overwrite of {}".format(target_path))

        with open(target_path, 'w') as f:
            f.write(payload)

    def __sort_qualifier(self, model_config):
        sort_keys = model_config['sort']
        if type(sort_keys) == str:
            sort_keys = [sort_keys]

        # remove existing quotes in field name, then wrap in quotes
        formatted_sort_keys = ['"{}"'.format(sort_key.replace('"', '')) for sort_key in sort_keys]
        return "sortkey ({})".format(', '.join(formatted_sort_keys))

    def __dist_qualifier(self, model_config):
        dist_key = model_config['dist']

        if type(dist_key) != str:
            raise RuntimeError("The provided distkey '{}' is not valid!".format(dist_key))

        return 'distkey ("{}")'.format(dist_key)

    def __wrap_in_create(self, model_name, query, model_config):

        # default to view if not provided in config!
        table_or_view = 'table' if model_config['materialized'] else 'view'

        ctx = self.project.context()
        schema = ctx['env'].get('schema', 'public')

        dist_qualifier = ""
        sort_qualifier = ""

        if table_or_view == 'table':
            if 'dist' in model_config:
                dist_qualifier = self.__dist_qualifier(model_config)
            if 'sort' in model_config:
                sort_qualifier = self.__sort_qualifier(model_config)

        opts = {
            "table_or_view": table_or_view,
            "schema": schema,
            "identifier": model_name,
            "query": query,
            "dist_qualifier": dist_qualifier,
            "sort_qualifier": sort_qualifier
        }

        return CREATE_STATEMENT_TEMPLATE.format(**opts)

    def __get_model_identifiers(self, model_filepath):
        model_group = os.path.dirname(model_filepath)
        model_name, _ = os.path.splitext(os.path.basename(model_filepath))
        return model_group, model_name

    def __get_model_config(self, model_group, model_name):
        """merges model, model group, and base configs together. Model config
        takes precedence, then model_group, then base config"""

        config = self.project['model-defaults'].copy()

        model_configs = self.project['models']
        model_group_config = model_configs.get(model_group, {})
        model_config = model_group_config.get(model_name, {})

        config.update(model_group_config)
        config.update(model_config)

        return config

    def __find_model_by_name(self, project_models, name, package_namespace=None):
        found = []
        for model in project_models:
            package, model_group, model_name = model
            if model_name == name:
                if package_namespace is None:
                    found.append(model)
                elif package_namespace is not None and package_namespace == package:
                    found.append(model)

        if len(found) == 0:
            raise RuntimeError("Can't find a model named '{}' in package '{}' -- does it exist?".format(name, package_namespace))
        elif len(found) == 1:
            return found[0]
        else:
            raise RuntimeError("Model specification is ambiguous: model='{}' package='{}' -- {} models match criteria".format(name, package_namespace, len(found)))

    def __ref(self, linker, ctx, source_model, project_models):
        schema = ctx['env']['schema']

        # if this node doesn't have any deps, still make sure it's a part of the graph
        linker.add_node(source_model)

        def do_ref(*args):
            if len(args) == 1:
                other_model_name = args[0]
                other_model = self.__find_model_by_name(project_models, other_model_name)
            elif len(args) == 2:
                other_model_package, other_model_name = args
                other_model = self.__find_model_by_name(project_models, other_model_name, package_namespace=other_model_package)
            linker.dependency(source_model, other_model)
            return '"{}"."{}"'.format(schema, other_model_name)

        return do_ref

    def __compile(self, src_index, project_models):
        linker = Linker()

        for src_path, project_files in src_index.items():
            jinja = jinja2.Environment(loader=jinja2.FileSystemLoader(searchpath=src_path))
            for (project, f) in project_files:

                model_group, model_name = self.__get_model_identifiers(f)
                model_config = self.__get_model_config(model_group, model_name)

                if not model_config.get('enabled'):
                    continue

                template = jinja.get_template(f)

                context = self.project.context()
                source_model = (project['name'], model_group, model_name)
                context['ref'] = self.__ref(linker, context, source_model, project_models)

                rendered = template.render(context)

                create_stmt = self.__wrap_in_create(model_name, rendered, model_config)

                if create_stmt:
                    self.__write(f, create_stmt)

        return linker

    def run(self):
        sources = self.__project_sources(self.project)

        for obj in os.listdir(self.project['modules-path']):
            full_obj = os.path.join(self.project['modules-path'], obj)
            if os.path.isdir(full_obj):
                project = dbt.project.read_project(os.path.join(full_obj, 'dbt_project.yml'))
                sources.update(self.__project_sources(project))

        project_models = self.__project_models(sources)
        linker = self.__compile(sources, project_models)

        graph_path = os.path.join(self.project['target-path'], 'graph.yml')
        linker.write_graph(graph_path)



import os
import errno
import re
import yaml
import pprint
import subprocess
import dbt.project as project

def folder_from_git_remote(remote_spec):
    start = remote_spec.rfind('/') + 1
    end = len(remote_spec) - (4 if remote_spec.endswith('.git') else 0)
    return remote_spec[start:end]

class DepsTask:
    def __init__(self, args, project):
        self.args = args
        self.project = project

    def __pull_repo(self, repo):
        proc = subprocess.Popen(
            ['git', 'clone', repo],
            cwd=self.project['modules-path'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)

        out, err = proc.communicate()

        exists = re.match("fatal: destination path '(.+)' already exists", err)
        folder = None
        if exists:
            folder = exists.group(1)
            print("updating existing dependency {}".format(folder))
            full_path = os.path.join(self.project['modules-path'], folder)
            proc = subprocess.Popen(
                ['git', 'fetch', '--all'],
                cwd=full_path,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
            out, err = proc.communicate()
            proc = subprocess.Popen(
                ['git', 'reset', '--hard', 'origin/master'],
                cwd=full_path,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
            out, err = proc.communicate()
        else:
            matches = re.match("Cloning into '(.+)'", err)
            folder = matches.group(1)
            print("pulled new dependency {}".format(folder))

        return folder

    def __pull_deps_recursive(self, repos, processed_repos = set()):
        for repo in repos:
            repo_folder = folder_from_git_remote(repo)
            try:
                if repo_folder in processed_repos:
                    print("skipping already processed dependency {}".format(repo_folder))
                else:
                    dep_folder = self.__pull_repo(repo)
                    dep_project = project.read_project(
                        os.path.join(self.project['modules-path'],
                                     dep_folder,
                                     'dbt_project.yml')
                    )
                    processed_repos.add(dep_folder)
                    self.__pull_deps_recursive(dep_project['repositories'], processed_repos)
            except IOError as e:
                if e.errno == errno.ENOENT:
                    print("'{}' is not a valid dbt project - dbt_project.yml not found".format(repo))
                    exit(1)
                else:
                    raise e

    def run(self):
        if not os.path.exists(self.project['modules-path']):
            os.makedirs(self.project['modules-path'])

        self.__pull_deps_recursive(self.project['repositories'])



import pprint
import psycopg2
import os, sys
import fnmatch

from dbt.compilation import Linker

class RedshiftTarget:
    def __init__(self, cfg):
        assert cfg['type'] == 'redshift'
        self.host = cfg['host']
        self.user = cfg['user']
        self.password = cfg['pass']
        self.port = cfg['port']
        self.dbname = cfg['dbname']
        self.schema = cfg['schema']

    def __get_spec(self):
        return "dbname='{}' user='{}' host='{}' password='{}' port='{}'".format(
            self.dbname,
            self.user,
            self.host,
            self.password,
            self.port
        )

    def get_handle(self):
        return psycopg2.connect(self.__get_spec())

class RunTask:
    def __init__(self, args, project):
        self.args = args
        self.project = project

        self.model_sql_map = {}

    def __compiled_files(self):
        compiled_files = []
        sql_path = self.project['target-path']

        for root, dirs, files in os.walk(sql_path):
            for filename in files:
                if fnmatch.fnmatch(filename, "*.sql"):
                    abs_path = os.path.join(root, filename)
                    rel_path = os.path.relpath(abs_path, sql_path)
                    compiled_files.append(rel_path)

        return compiled_files

    def __get_target(self):
        target_cfg = self.project.run_environment()
        if target_cfg['type'] == 'redshift':
            return RedshiftTarget(target_cfg)
        else:
            raise NotImplementedError("Unknown target type '{}'".format(target_cfg['type']))

    def __deserialize_graph(self, linker):
        graph_file = os.path.join(self.project['target-path'], 'graph.yml')
        linker.read_graph(graph_file)

    def __create_schema(self):
        target_cfg = self.project.run_environment()
        target = self.__get_target()
        with target.get_handle() as handle:
            with handle.cursor() as cursor:
                cursor.execute('create schema if not exists "{}"'.format(target_cfg['schema']))

    def __load_models(self):
        target = self.__get_target()
        for f in self.__compiled_files():
            with open(os.path.join(self.project['target-path'], f), 'r') as fh:
                namespace = os.path.dirname(f)
                model_name, _ = os.path.splitext(os.path.basename(f))

                model = (self.project['name'], namespace, model_name)
                self.model_sql_map[model] = fh.read()

    def __query_for_existing(self, cursor, schema):
        sql = """
            select tablename as name, 'table' as type from pg_tables where schemaname = '{schema}'
                union all
            select viewname as name, 'view' as type from pg_views where schemaname = '{schema}' """.format(schema=schema)

        cursor.execute(sql)
        existing = [(name, relation_type) for (name, relation_type) in cursor.fetchall()]

        return dict(existing)

    def __drop(self, cursor, schema, relation, relation_type):
        sql = 'drop {relation_type} if exists "{schema}"."{relation}" cascade'.format(schema=schema, relation_type=relation_type, relation=relation)
        cursor.execute(sql)

    def __execute_models(self, linker):
        target = self.__get_target()

        with target.get_handle() as handle:
            with handle.cursor() as cursor:

                existing = self.__query_for_existing(cursor, target.schema);

                dependency_list = list(linker.as_dependency_list())

                if len(dependency_list) == 0:
                    print("WARNING: Target directory is empty: '{}'. Try running `dbt compile`.".format(self.project['target-path']))
                    return

                for model in dependency_list:
                    package_name, namespace, model_name = model
                    if model_name in existing:
                        print("dropping {} '{}.{}'".format(existing[model_name], target.schema, model_name))
                        self.__drop(cursor, target.schema, model_name, existing[model_name])
                        handle.commit()

                    print("creating {}".format(model_name))
                    sql = self.model_sql_map[model]
                    cursor.execute(sql)
                    print("         {}".format(cursor.statusmessage))
                    handle.commit()

    def run(self):
        linker = Linker()
        self.__deserialize_graph(linker)
        self.__load_models()

        try:
            self.__create_schema()
            self.__execute_models(linker)
        except psycopg2.OperationalError as e:
            print("ERROR: Could not connect to the target database. Try `dbt debug` for more information")
            print(e.message)
            sys.exit(1)



import os.path
import yaml
import pprint
import copy
import sys

default_project_cfg = {
    'source-paths': ['models'],
    'test-paths': ['test'],
    'target-path': 'target',
    'clean-targets': ['target'],
    'outputs': {'default': {}},
    'run-target': 'default',
    'models': {},
    'model-defaults': {
        'enabled': True,
        'materialized': False
    },
    'repositories': [],
    'modules-path': 'dbt_modules'
}

default_profiles = {
    'user': {}
}

default_active_profiles = ['user']

class DbtProjectError(Exception):
    def __init__(self, message, project):
        self.project = project
        super(DbtProjectError, self).__init__(message)

class Project:

    def __init__(self, cfg, profiles, active_profile_names=[]):
        self.cfg = default_project_cfg.copy()
        self.cfg.update(cfg)
        self.profiles = default_profiles.copy()
        self.profiles.update(profiles)
        self.active_profile_names = active_profile_names

        for profile_name in active_profile_names:
            self.cfg.update(self.profiles[profile_name])

    def __str__(self):
        return pprint.pformat({'project': self.cfg, 'profiles': self.profiles})

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, key):
        return self.cfg.__getitem__(key)

    def __contains__(self, key):
        return self.cfg.__contains__(key)

    def __setitem__(self, key, value):
        return self.cfg.__setitem__(key, value)

    def get(self, key, default=None):
        return self.cfg.get(key, default)

    def run_environment(self):
        target_name = self.cfg['run-target']
        return self.cfg['outputs'][target_name]

    def context(self):
        target_cfg = self.run_environment()
        filtered_target = copy.deepcopy(target_cfg)
        filtered_target.pop('pass', None)
        return {'env': filtered_target}

    def with_profiles(self, profiles=[]):
        return Project(
            copy.deepcopy(self.cfg),
            copy.deepcopy(self.profiles),
            profiles)

    def validate(self):
        target_cfg = self.run_environment()
        target_name = self.cfg['run-target']

        package_name = self.cfg.get('name', None)
        package_version = self.cfg.get('version', None)

        if package_name is None or package_version is None:
            raise DbtProjectError("Project name and version is not provided", self)

        required_keys = ['host', 'user', 'pass', 'schema', 'type', 'dbname', 'port']
        for key in required_keys:
            if key not in target_cfg or len(str(target_cfg[key])) == 0:
                raise DbtProjectError("Expected project configuration '{}' was not supplied".format(key), self)

def read_profiles():
    profiles = {}
    paths = [
        os.path.join(os.path.expanduser('~'), '.dbt/profiles.yml')
    ]
    for path in paths:
        if os.path.isfile(path):
            with open(path, 'r') as f:
                m = yaml.safe_load(f)
                profiles.update(m)

    return profiles

def read_project(filename):
    with open(filename, 'r') as f:
        project_cfg = yaml.safe_load(f)
        project_cfg['project-root'] = os.path.dirname(os.path.abspath(filename))
        profiles = read_profiles()
        proj = Project(project_cfg, profiles, default_active_profiles)

        proj.validate()
        return proj



#!/usr/bin/env python
from setuptools import setup, find_packages
import os.path

package_name = "dbt"
package_version = "0.1.14"

setup(
  name=package_name,
  version=package_version,
  description="Data build tool for Analyst Collective",
  author="Analyst Collective",
  author_email="admin@analystcollective.org",
  url="https://github.com/analyst-collective/dbt",
  packages=find_packages(),
  entry_points={
    'console_scripts': [
      'dbt = dbt.main:main',
    ],
  },
  scripts=[
    'scripts/dbt',
  ],
  install_requires=[
    'Jinja2>=2.8',
    'PyYAML>=3.11',
    'psycopg2==2.6.1',
    'sqlparse==0.1.19',
    'networkx==1.11',
  ],
)



import os.path
import os
import shutil


class CleanTask:

    def __init__(self, args, project):
        self.args = args
        self.project = project

    def __is_project_path(self, path):
        proj_path = os.path.abspath('.')
        return not os.path.commonprefix(
            [proj_path, os.path.abspath(path)]
        ) == proj_path

    def __is_protected_path(self, path):
        abs_path = os.path.abspath(path)
        protected_paths = self.project['source-paths'] + self.project['test-paths'] + ['.']

        protected_abs_paths = [os.path.abspath for p in protected_paths]
        return abs_path in set(protected_abs_paths) or self.__is_project_path(abs_path)

    def run(self):
        for path in self.project['clean-targets']:
            if not self.__is_protected_path(path):
                shutil.rmtree(path, True)




import networkx as nx

class Linker(object):
    def __init__(self):
        self.graph = nx.DiGraph()

    def as_dependency_list(self):
        return nx.topological_sort(self.graph, reverse=True)

    def dependency(self, node1, node2):
        "indicate that node1 depends on node2"
        self.graph.add_node(node1)
        self.graph.add_node(node2)
        self.graph.add_edge(node1, node2)

    def add_node(self, node):
        self.graph.add_node(node)

    def write_graph(self, outfile):
        nx.write_yaml(self.graph, outfile)

    def read_graph(self, infile):
        self.graph = nx.read_yaml(infile)




import argparse
import os.path
import sys
import dbt.project as project
import dbt.task.run as run_task
import dbt.task.compile as compile_task
import dbt.task.debug as debug_task
import dbt.task.clean as clean_task
import dbt.task.deps as deps_task

def main(args=None):
    if args is None:
        args = sys.argv[1:]

    if os.path.isfile('dbt_project.yml'):
        proj = project.read_project('dbt_project.yml')
    else:
        raise RuntimeError("dbt must be run from a project root directory with a dbt_project.yml file")

    p = argparse.ArgumentParser(prog='dbt: data build tool')
    subs = p.add_subparsers()

    base_subparser = argparse.ArgumentParser(add_help=False)
    base_subparser.add_argument('--profile', default=["user"], nargs='+', type=str, help='Which profile to load')

    sub = subs.add_parser('clean', parents=[base_subparser])
    sub.set_defaults(cls=clean_task.CleanTask)

    sub = subs.add_parser('compile', parents=[base_subparser])
    sub.set_defaults(cls=compile_task.CompileTask)

    sub = subs.add_parser('debug', parents=[base_subparser])
    sub.set_defaults(cls=debug_task.DebugTask)

    sub = subs.add_parser('deps', parents=[base_subparser])
    sub.set_defaults(cls=deps_task.DepsTask)

    sub = subs.add_parser('run', parents=[base_subparser])
    sub.set_defaults(cls=run_task.RunTask)

    parsed = p.parse_args(args)

    proj = proj.with_profiles(parsed.profile)
    parsed.cls(args=parsed, project=proj).run()
